classdef VhistFile
% VHISTFILE    Holds information about one file in a VHIST section.
%    All information about one file inside of a VHIST workflowstep are
%    stored in this class. Most fields can be set to an empty cell array {}
%    to use the default value.
%
% VHISTFILE Properties:
%    filename        - The name of the file.
%    purpose         - Whether the file was an input or an output of the
%                      documented command.
%    description     - A description of the files content.
%    comment         - Additional comments.
%    filetype        - The MIME type of the file.
%    md5file         - The path to a file containing the md5sum of the
%                      embedded file.
%    previews        - A cell array of preview images or figures of this
%                      file.
%    isPreviewWs     - Whether or not the file is a preview image of the
%                      generated workflowstep.
%    isPreview       - Whether or not the file is a preview image of the
%                      previous embedded file in the workflow step.
%    isThumbnail     - Whether or not the file is a thumbnail image for the
%                      filetype of the previous embedded file in the
%                      worklfow step.
%    isThumbnailOnly - Same as isThumbnail, but image will not appear in
%                      the summary inside the generated VHIST file.
%    optional        - Whether or not the file must exist in the
%                      filesystem.
%    automd5         - Whether or not the md5 checksum of the file is
%                      automatically generated by vhistadd.
%    embed           - Whether or not the file's content should be embedded
%                      inside the generated VHIST file.
%    compress        - Whether or not the file's content should be
%                      embedded in compressed form.
%    userAttrs       - A containters.Map object with strings as keys and
%                      values for arbitrary, user-defined key-value-pairs.

    properties
        % FILENAME    The filename of the embedded file.
        %    Should including the filepath (absolute or relative to the
        %    current working directory).
        filename = {};
        % PURPOSE    The purpose of the embedded file.
        %    Must be either 'in' for a file which was an input file in the
        %    documented workflowstep or 'out' for a file which was an
        %    output file.
        purpose = {};
        % DESCRIPTION    A short description of the embedded file.
        %    Can either be a string or an empty cell array {} if no
        %    description should be supplied.
        description = {};
        % COMMENT    A freeform comment regarding the file.
        %    This comment can contain any addition information about the
        %    embedded file. May be an empty cell array {} if no comment is
        %    neccessary.
        comment = {};
        % FILETYPE    A type of the file.
        %    The filetype should follows the conventions of the MIME
        %    standard, e.g. image/jpeg or application/pdf. An empty cell {}
        %    can be used to provide no filetype.
        filetype = {};
        % MD5FILE    A file containing the md5 checksum of this file.
        %    This property is useful only if the file should not be
        %    embedded and the md5sum has already been generated by another
        %    program. In this case, set embed and automd5 to false.
        %    Filetype must point to a file which only contains the md5
        %    checksum for this file. An empty cell array {} can be used to
        %    ignore this option.
        %    See also VhistFile.embed and VhistFile.automd5
        md5file = {};
        % PREVIEWS    A cell array of previews for this file.
        %    Each entry in the cell array can be either the filepath to a
        %    png file, the handle to a figure or a 2d- or 3d-matrix, which
        %    can be saved using the imwrite function.
        %    See also figure and imwrite.
        previews = {};
        % ISPREVIEWWS    Whether or not the file is a preview image of the
        %                generated workflowstep.
        %    If this flag is set to true, the file will be interpreted as
        %    a preview image for the generated workflowset. Usually, you
        %    should not use this option by yourself. Instead, use the
        %    VhistSection.previews property. Ignore this option by
        %    supplying an empty cell array {}.
        %    See also VhistSection.previews.
        isPreviewWs = {};
        % ISPREVIEW    Whether or not the file is a preview image of the
        %              previous file in the workflow step.
        %    If this flag is set to true, the file will be interpreted as
        %    a preview image for the preview embedded file in the
        %    workflowstep, i.e. VhistSection.files. Usually, you should not
        %    use this option by yourself. Instead, use the
        %    VhistFile.previews property. The default value can be used by
        %    setting this property to an empty cell array {}.
        %    See also VhistFile.previews.
        isPreview = {};
        isThumbnail = {};
        isThumbnailOnly = {};
        % OPTIONAL    Whether the file can be ignored by VHIST.
        %    In the case that the optional flag is set to false,
        %    VhistFile.embed or VhistFile.automd5 is set and the file
        %    does not exist in the file system, VHIST will abort with an
        %    error message. If the optional flag, however, is set to true,
        %    the file will simply be ignored. Useful for files which may
        %    or may not exist. If this option is set to an empty cell
        %    array {}, the default value will be used.
        %    See also VhistFile.embed and VhistFile.automd5.
        optional = {};
        % AUTOMD5    Whether an md5 checksum of the file should be
        %            automatically generated or not.
        %    Usually, VHIST automatically generates a md5 checksum for
        %    every file. For very large files, however, calculating the
        %    checksum takes a large amount of time and it can be
        %    appropriate to deactivate the generation of the checksum using
        %    this option. If the VhistFile.embed flag is set to true, a
        %    checksum will always be generated and this option is ignored.
        %    See also VhistFile.embed and VhistFile.optional.
        automd5 = {};
        % EMBED    Whether or not the file should be embedded.
        %     If a file is embedded into the VHIST file, it's content will
        %     be stored inside the VHIST file and can be extracted using
        %     most PDF viewers or VHIST specific tools. Most often, it is
        %     suitable to embed all but very large files into a VHIST file.
        %     An empty cell array {} can be supplied to use the default
        %     setting.
        %     See also VhistFile.automd5 and VhistFile.optional.
        embed = {};
        % COMPRESS    Whether or not an embedded file should be stored in
        %             compressed form.
        %    The file is compressed using the flate compression algorithm.
        %    Often, the size of a file can be reduced significantly by
        %    compressing it, however, if the file is already compressed, it
        %    is useful to deactivate compression to decrease computation
        %    time. An empty cell array {} can be supplied to use the
        %    default value.
        %    See also VhistFile.embed.
        compress = {};
        
        % USERATTRS    A container.Map object for user-defined information.
        %    Both, keys and values of the container.Map object must be
        %    strings. The purpose of this dictionary is to store additional
        %    information, e.g. parameter settings, patient ids or used
        %    methods.
        userAttrs = containers.Map;
        private__previewFiles = [];

    end
    
    methods
        function obj = VhistFile(filename, purpose, description, filetype, ...
                comment, embed, compress, optional, automd5, md5file, previewWs, ...
                preview, thumbnail, thumbnailOnly)
            % VHISTFILE    Constructor of the VhistFile class.
            %    This method creates a new VHIST embedded file. It sets the
            %    values of the most commonly used properties of this class.
            %    
            %    FILE = VHISTFILE(FILENAME, PURPOSE, DESCRIPTION, ...
            %           FILETYPE, COMMENT, EMBED, COMPRESS, OPTIONAL, ...
            %           AUTOMD5, MD5FILE, PREVIEWWS, PREVIEW, ...
            %           THUMBNAIL, THUMBNAILONLY)
            %    All parameters exept for FILENAME and PURPOSE are
            %    optional.
            %    FILENAME sets the value of the property filename.
            %    PURPOSE sets the value of the property purpose.
            %    DESCRIPTION sets the value of the property description.
            %    FILETYPE sets the value of the property filetype.
            %    COMMENT sets the value of the property comment.
            %    EMBED sets the value of the property embed.
            %    COMPRESS sets the value of the property compress.
            %    OPTIONAL sets the value of the property optional.
            %    AUTOMD5 sets the value of the property automd5.
            %    MD5FILE sets the value of the property md5file.
            %    PREVIEWWS sets the value of the property isPreviewWs.
            %    PREVIEW sets the value of the property isPreview.
            %    THUMBNAIL sets the value of the property isThumbnail.
            %    THUMBNAILONLY sets the value of the property
            %    isThumbnailOnly.
            %    See also VhistFile.filename, VhistFile.purpose,
            %    VhistFile.description, VhistFile.filetype,
            %    VhistFile.comment, VhistFile.embed, VhistFile.compress,
            %    VhistFile.optional, VhistFile.automd5, VhistFile.md5file,
            %    VhistFile.isPreviewWs, VhistFile.isPreview,
            %    VhistFile.isThumbnail and VhistFile.isThumbnailOnly.
            
            obj.filename = filename;
            obj.purpose = purpose;
            
            if nargin >= 3
                obj.description = description;
            end
            if nargin >= 4
                obj.filetype = filetype;
            else
                obj.filetype = matvhist.getMimeType(obj.filename);
            end
            if nargin >= 5
                obj.comment = comment;
            end
            if nargin >= 6
                obj.embed = embed;
            end
            if nargin >= 7
                obj.compress = compress;
            end
            if nargin >= 8
                obj.optional = optional;
            end
            if nargin >= 9
                obj.automd5 = automd5;
            end
            if nargin >= 10
                obj.md5file = md5file;
            end
            if nargin >= 11
                obj.isPreviewWs = previewWs;
            end
            if nargin >= 12
                obj.isPreview = preview;
            end
            if nargin >= 13
                obj.isThumbnail = thumbnail;
            end
            if nargin >= 14
                obj.isThumbnailOnly = thumbnailOnly;
            end
            obj.userAttrs = containers.Map();
        end
        
        
        function obj = set.filename(obj, value)
            if ischar(value)
                obj.filename = value;
            else
                throw(wrongTypeExceptionStringOnly('filename'));
            end
        end
        
        function obj = set.purpose(obj, value)
            if ischar(value) && (strcmp(value, 'in') || strcmp(value, 'out'))
                obj.purpose = value;
            else
                throw(wrongTypeExceptionPurpose())
            end
        end

        function obj = set.description(obj, value)
            if ischar(value) || (iscell(value) || isempty(value))
                obj.description = value;
            else
                throw(wrongTypeException('description'));
            end
        end
        
        function obj = set.filetype(obj, value)
            if ischar(value) || (iscell(value) || isempty(value))
                obj.filetype = value;
            else
                throw(wrongTypeException('description'));
            end
        end

        function obj = set.comment(obj, value)
            if ischar(value) || (iscell(value) || isempty(value))
                obj.comment = value;
            else
                throw(wrongTypeException('comment'));
            end
        end

        function obj = set.embed(obj, value)
            if islogical(value) || (iscell(value) || isempty(value))
                obj.embed = value;
            else
                throw(wrongTypeExceptionBool('embed'));
            end
        end
        
        function obj = set.compress(obj, value)
            if islogical(value) || (iscell(value) || isempty(value))
                obj.compress = value;
            else
                throw(wrongTypeExceptionBool('embed'));
            end
        end
        
        function obj = set.md5file(obj, value)
            if ischar(value) || (iscell(value) || isempty(value))
                obj.md5file = value;
            else
                throw(wrongTypeException('comment'));
            end
        end
        
        function obj = set.optional(obj, value)
            if islogical(value) || (iscell(value) || isempty(value))
                obj.optional = value;
            else
                throw(wrongTypeExceptionBool('embed'));
            end
        end
        
        function obj = set.automd5(obj, value)
            if islogical(value) || (iscell(value) || isempty(value))
                obj.automd5 = value;
            else
                throw(wrongTypeExceptionBool('embed'));
            end
        end
        
        function obj = set.isPreviewWs(obj, value)
            if islogical(value) || (iscell(value) || isempty(value))
                obj.isPreviewWs = value;
            else
                throw(wrongTypeExceptionBool('embed'));
            end
        end

        function obj = set.isPreview(obj, value)
            if islogical(value) || (iscell(value) || isempty(value))
                obj.isPreview = value;
            else
                throw(wrongTypeExceptionBool('embed'));
            end
        end

        function obj = set.isThumbnail(obj, value)
            if islogical(value) || (iscell(value) || isempty(value))
                obj.isThumbnail = value;
            else
                throw(wrongTypeExceptionBool('embed'));
            end
        end
        
        function obj = set.isThumbnailOnly(obj, value)
            if islogical(value) || (iscell(value) || isempty(value))
                obj.isThumbnailOnly = value;
            else
                throw(wrongTypeExceptionBool('embed'));
            end
        end

        function obj = set.userAttrs(obj, value)
            if strcmp(class(value), 'containers.Map')
                obj.userAttrs = value;
            else
                throw(wrongTypeExceptionMap('userAttrs'));
            end
        end
    end
end

% exceptions

function [except] = wrongTypeException(propname)
    str = sprintf('''%s'' must either be set to a string or an empty cell', propname);
    except = MException('vhist:wrongType', str);
end

function [except] = wrongTypeExceptionStringOnly(propname)
    str = sprintf('''%s'' must be set to a string', propname);
    except = MException('vhist:wrongType', str);
end

function [except] = wrongTypeExceptionPurpose()
    str = '''purpose'' must be either ''in'' or ''out''';
    except = MException('vhist:wrongType', str);
end

function [except] = wrongTypeExceptionBool(propname)
    str = sprintf('''%s'' must be either true or false or an empty cell', propname);
    except = MException('vhist:wrongType', str);
end

function [except] = wrongTypeExceptionMap(propname)
    str = sprintf('''%s'' must be a map with keys of type char', propname);
    except = MException('vhist:wrongType', str);
end
