classdef VhistRecorder < handle
% VHISTRECORDER    Records information about for a VHIST workflowstep.
%    Records various information about a workflowstep and makes them
%    available to VHIST. VhistRecorder records the start- end endtime of
%    the workflowstep, the duration, the used CPU time as well as a log
%    containing all information printed to the command window by the matlab
%    session. The information can be added to a VhistSection object.
%    Recording starts the moment, the startRecording method is called and
%    stopped when stopRecording is called.
%
%    VHISTRECORDER methods:
%    startRecording          - Starts recording of information.
%    stopRecording           - Stops recording of information.
%    addInformationToSection - Adds the recorded information to a
%                              VhistSection object.

    properties
        tmpDir = '';
        diaryFile = '';
        cpuStartTime = 0;
        cpuEndTime = 0;
        realStartTime = 0;
        realEndTime = 0;
        diaryDeleter = {};
    end
    
    methods
        function startRecording(obj)
            % STARTRECORDING    Starts recording information.
            %    The log file will be stored in a temporary file.
            %    startRecording uses matlabs diary function, which only
            %    supports recording to one diary. Therefore, calling this
            %    method multiple times or using diary and VhistRecorder in
            %    parallel will generate unexpected behaviour.
            
            % create diary
            obj.tmpDir = tempname();
            mkdir(obj.tmpDir);
            obj.diaryDeleter = matvhist.TempFile(obj.tmpDir);
            obj.diaryFile = [obj.tmpDir filesep 'matlab-diary-log.txt'];
            diary(obj.diaryFile);
            diary on;
            
            % measure time
            obj.cpuStartTime = cputime();
            obj.realStartTime = clock();
         end

        function stopRecording(obj)
            % STOPRECORDING    Stops recording information.
            %    Stops recording the log file and determines th ending time
            %    of the workflowstep.
            diary off;
            obj.cpuEndTime = cputime();
            obj.realEndTime = clock();
        end
        
        function [modifiedSection] = addInformationToSection(obj, vhistSection)
            % ADDINFORMATIONTOSECTION    Adds recorded information to a
            %                            section.
            %    Creates a new section from an existing one. The new
            %    section contains the recorded information as user-defined
            %    key-value pairs. Do not call this method before you have
            %    called startRecording and endRecording.
            %
            %    [NEWSEC] = RECORDER.ADDINFORMATIONTOSECTION[OLDSEC]
            %
            %    Adds the recorded information to an existing section and
            %    returns the result as a new section. OLDSEC will not be
            %    changed.
            
            modifiedSection = vhistSection;
            modifiedSection.userAttrs = matvhist.copyMap(vhistSection.userAttrs);
            
            elapsedTime = etime(obj.realEndTime, obj.realStartTime);
            elapsedCpuTime = obj.cpuEndTime - obj.cpuStartTime;
            elapsedTimeStr = [num2str(elapsedTime) ' secs'];
            elapsedCpuTimeStr = [num2str(elapsedCpuTime) ' secs'];
            startTime = datestr(obj.realStartTime, 'yyyy-mm-dd HH:MM:SS');
            endTime = datestr(obj.realEndTime, 'yyyy-mm-dd HH:MM:SS');
            
            modifiedSection.userAttrs('timing.elapsedRealTime') = elapsedTimeStr;
            modifiedSection.userAttrs('timing.elapsedCpuTime') = elapsedCpuTimeStr;
            modifiedSection.userAttrs('timing.startTime') = startTime;
            modifiedSection.userAttrs('timing.endTime') = endTime;
            
            logfile = matvhist.VhistFile(obj.diaryFile, 'out');
            logfile.description = 'The MATLAB log-diary-file';
            modifiedSection.files = [modifiedSection.files logfile];
            if ~isempty(obj.tmpDir)
                modifiedSection.private__tempFiles = ...
                        [modifiedSection.private__tempFiles, {obj.diaryDeleter}];
            end
        end
    end
end
