classdef VhistSection
% VHISTSECTION Holds information about one workflowstep in a VHIST file.
%    All information about one VHIST workflowstep are
%    stored in this class. Along with a set of general information,
%    arbitrary case-specific information as well as a list of embedded
%    files can be included with the workflowstep. Most properties can be
%    set to an empty cell array {} to use the default value of the
%    property.
%
% VHISTSECTION Properties:
%    title       - The title of a workflowstep.
%    description - A short description stating the action performed in the
%                  workflowstep.
%    comment     - An comment with additional information about the
%                  workflowstep.
%    command     - The command executed in this workflowstep, e.g. a
%                  commandline call or a function call inside a programming
%                  language.
%    host        - The name of the computer, on which the command was
%                  executed.
%    user        - The user, who executed the command.
%    tool        - The program or script, which was executed.
%    toolpath    - The path of the programs executable.
%    previews    - A cell array of preview images or figures of this
%                  workflowstep.
%    userAttrs   - A containers.Map object containing user-defined
%                  key-value-pairs.
%    files       - A vector of VhistFile objects for files, which were used
%                  as inputs or outputs in the workflowstep.

    properties
        % TITLE    The title of the workflowstep. 
        %    The title of a workflowstep should be a rather short string.
        %    More information can be stored inside the description and
        %    comment properties. An empty cell array {} can be used to
        %    provide no title.
        %    See also VhistSection.description and VhistSection.comment.
        title = {};
        % DESCRIPTION    A short description of the workflowstep.
        %    The description should be between one and three sentenses
        %    long. If you want to include a longer description, e.g. the
        %    abstract of a paper or the paper itself, it is better to
        %    include this information as an embedded file. An empty cell
        %    array {} can be used to disable the description.
        description = {};
        % COMMENT    A comment containing additional information about the
        %            workflowstep. No comment will be generated if an empty
        %            cell array {} is used as the value for this property.
        comment = {};
        % COMMAND    The command executed in this workflowstep.
        %            Usually, this will be the name of the matlab script or
        %            function used. An empty cell array can be used to
        %            supply no command.
        command = {};
        % HOST    The name of the computer, on which this workflowstep has
        %         been performed.
        %    If no hostname should be supplied, use an empty cell array {}.
        host = {};
        % USER    The name of the user, who executed this workflowstep.
        %    If no username should be provided or the user is not known,
        %    use an empty cell array {}.
        user = {};
        % TOOL    The name of the tool or program, which was used to create
        %         this workflowstep.
        %    An empty cell array {} can be used to not provide a tool.
        tool = {};
        % TOOLPATH The path to the tool's executable inside the filesystem.
        %    An empty cell array {} can be used to omit the toolpath.
        toolpath = {};
        % PREVIEWS    A cell array of previews for this workflowstep.
        %    Each entry in the cell array can be either the filepath to a
        %    png file, the handle to a figure or a 2d- or 3d-matrix, which
        %    can be saved using the imwrite function.
        %    See also figure and imwrite.
        previews = {};
        private__previewFiles = [];
        private__tempFiles = {};
        % USERATTRS    A container.Map object for user-defined information.
        %    Both, keys and values of the container.Map object must be
        %    strings. The purpose of this dictionary is to store additional
        %    information, e.g. parameter settings, patient ids or used
        %    methods.
        userAttrs = containers.Map;
        % FILES    A vector of VhistFile objects of embedded files.
        %    All files involved in the workflowstep can be included into it
        %    by adding them to this property. For more information about
        %    these files, see the documentation of the VhistFile class.
        %    See also VhistFile.
        files = [];
    end
    
    methods
        function obj = VhistSection(title, description, comment, command, host, user, tool, toolpath)
            % VHISTSECTION Constructor of the VhistSection class.
            %    This method creates a new VHIST section, also called a
            %    workflowstep. This constructor simply initializes the
            %    values of the most commonly used properties of this class.
            %    
            %    SECTION = VHISTSECTION(TITLE, DESCRIPTION, COMMENT, ...
            %              COMMAND, HOST, USER, TOOL, TOOLPATH)
            %    All parameters are optional.
            %    TITLE sets the value of the property title.
            %    DESCRIPTION sets the value of the property description.
            %    COMMENT sets the value of the property comment.
            %    COMMAND sets the value of the property command.
            %    HOST sets the value of the property host.
            %    USER sets the value of the property user.
            %    TOOL sets the value of the property tool.
            %    TOOLPATH sets the value of the property toolpath.
            %    The result is VhistSection class instance.
            %    
            %    The properties host, user, tool and toolpath are
            %    automatically set to suitable values by the constructor if
            %    they are omitted by the caller, so usually, it is not
            %    neccessary to suply values for these parameters.
            %    See also VhistSection.title, VhistSection.description,
            %    VhistSection.comment, VhistSection.command,
            %    VhistSection.host, VhistSection.user, VhistSection.tool
            %    and VhistSection.toolpath.
            if nargin >= 1
                obj.title = title;
            end
            if nargin >= 2
                obj.description = description;
            end
            if nargin >= 3
                obj.comment = comment;
            end
            if nargin >= 4
                obj.command = command;
            end
            if nargin >= 5
                obj.host = host;
            else
                obj.host = matvhist.getHostname();
            end
            if nargin >= 6
                obj.user = user;
            else
                obj.user = matvhist.getUsername();
            end
            if nargin >= 7
                obj.tool = tool;
            else
                obj.tool = ['Matlab ' version];
            end
            if nargin >= 8
                obj.toolpath = toolpath;
            else
                obj.toolpath = matlabroot;
            end
            obj.userAttrs = containers.Map();
        end
        
        
        function obj = set.title(obj, value)
            if ischar(value) || (iscell(value) || isempty(value))
                obj.title = value;
            else
                throw(wrongTypeException('title'));
            end
        end
        
        
        function obj = set.description(obj, value)
            if ischar(value) || (iscell(value) || isempty(value))
                obj.description = value;
            else
                throw(wrongTypeException('description'));
            end
        end
        
        
        function obj = set.comment(obj, value)
            if ischar(value) || (iscell(value) || isempty(value))
                obj.comment = value;
            else
                throw(wrongTypeException('comment'));
            end
        end
        
        
        function obj = set.command(obj, value)
            if ischar(value) || (iscell(value) || isempty(value))
                obj.command = value;
            else
                throw(wrongTypeException('command'));
            end
        end
        
        
        function obj = set.host(obj, value)
            if ischar(value) || (iscell(value) || isempty(value))
                obj.host = value;
            else
                throw(wrongTypeException('host'));
            end
        end
        
        
        function obj = set.user(obj, value)
            if ischar(value) || (iscell(value) || isempty(value))
                obj.user = value;
            else
                throw(wrongTypeException('user'));
            end
        end
        
        
        function obj = set.tool(obj, value)
            if ischar(value) || (iscell(value) || isempty(value))
                obj.tool = value;
            else
                throw(wrongTypeException('tool'));
            end
        end
        
        
        function obj = set.toolpath(obj, value)
            if ischar(value) || (iscell(value) || isempty(value))
                obj.toolpath = value;
            else
                throw(wrongTypeException('toolpath'));
            end
        end
        
        
        function obj = set.userAttrs(obj, value)
            if strcmp(class(value), 'containers.Map')
                obj.userAttrs = value;
            else
                throw(wrongTypeExceptionMap('userAttrs'));
            end
        end
        
        
        function obj = set.previews(obj, value)
            if iscell(value)
                allAreValid = true;
                for i=1:numel(value)
                    curElem = value{i};
                    allAreValid = allAreValid && (isValidString(curElem) || isValidFigure(curElem) || isValidImage(curElem));
                end
                if allAreValid
                    obj.previews = value;
                else
                    throw(wrongTypeExceptionPreviews('previews'));
                end
            end
        end
        
        
        function obj = set.files(obj, value)
            if isempty(value) || strcmp(class(value), 'matvhist.VhistFile')
                obj.files = value;
            else
                throw(wrongTypeExceptionFile('files'));
            end
        end

        
        function [result, stdout] = appendSectionToVhistFile(obj, ...
                vhistFile, vhistExecutable)
            
            if nargin < 3
                vhistExecutable = matvhist.getVhistExec();
            end
            
            prefix = vhistFilesToPrefix(vhistFile, {}, '-A', ...
                    '-O', false, 'vhistFile');
            
            [result, stdout] = obj.callVhistadd(prefix, vhistExecutable);
        end
        
        
        function [result, stdout] = addVhistSection(obj, ...
                outputVhistFiles, inputVhistFile, vhistExecutable)
            
            if nargin < 3
                inputVhistFile = '';
            end
            
            if nargin < 4
                vhistExecutable = matvhist.getVhistExec();
            end
            
            if ~isempty(inputVhistFile)
                prefix = {'-J', inputVhistFile};
            else
                prefix = {};
            end
            
            prefix = vhistFilesToPrefix(outputVhistFiles, prefix, ...
                '-O', '-O', false, 'outputVhistFile');
            
            [result, stdout] = obj.callVhistadd(prefix, vhistExecutable);
        end
        
        
        function [result, stdout] = callVhistadd(obj, prefix, ...
                vhistExecutable)
            
            tmpDir = tempname();
            mkdir(tmpDir);
            tmpDeleter = matvhist.TempFile(tmpDir);
            
            % create previews
            sectionWithPreviews = matvhist.createPreviews(obj, tmpDir);
            
            for i = 1:numel(sectionWithPreviews.files)
                curFile = sectionWithPreviews.files(i);
                sectionWithPreviews.files(i) = matvhist.createPreviews(curFile, tmpDir);
            end
            
            % generate argfile
            argfileContent = matvhist.buildArgfileText(sectionWithPreviews, ...
                    prefix, {});
            
            argfile = tempname();
            fid = fopen(argfile, 'w', 'native', 'UTF-8');
            tmpDeleterArg = matvhist.TempFile(argfile);
            fwrite(fid, argfileContent);
            fclose(fid);
            
            % run vhistadd
            [result, stdout] = system([vhistExecutable ' -c ' argfile]);
            if result ~= 0
                throw(matvhist.VhistaddException(result, stdout));
            end
        end
    end
end


function newPrefix = vhistFilesToPrefix(vhistFiles, oldPrefix, ...
        firstCommand, restCommands, canBeEmpty, exceptionVariableName)
    
    newPrefix = oldPrefix;
    if ischar(vhistFiles)
        newPrefix = [newPrefix, firstCommand, vhistFiles];
    elseif isValidCellstr(vhistFiles) && ~isempty(vhistFiles)
        newPrefix = [newPrefix, firstCommand, vhistFiles{1}];
        for i = 2 : numel(vhistFiles)
            newPrefix = [newPrefix, restCommands, vhistFiles{i}];
        end
    elseif iscell(vhistFiles) && isempty(vhistFiles) && canBeEmpty
        newPrefix = oldPrefix;
    else
        if canBeEmpty
            throw(MException('vhist:wrongType', ...
                '%s must be either a string or a cellstr.', ...
                exceptionVariableName));
        else
            throw(MException('vhist:wrongType', ...
                ['%s must be either a string or a cellstr, which ' ...
                'contains at least one string.'], exceptionVariableName));
        end
    end
end


function [validCellString] = isValidCellstr(scell)

validCellString = iscell(scell) && all(cellfun(@ischar, scell));

end


function [validFigure] = isValidFigure(fid)
    validFigure = isnumeric(fid) && numel(fid) == 1;
end


function [validString] = isValidString(str)
    validString = ischar(str);
end


function [validImage] = isValidImage(img)
    if ndims(img) <= 2
        validImage = true;
    elseif ndims(img) == 3 && size(img, 3) == 3
        validImage = true;
    else
        validImage = false;
    end
end


function [except] = wrongTypeException(propname)
    str = sprintf('''%s'' must either be set to a string or an empty cell', propname);
    except = MException('matvhist:wrongType', str);
end


function [except] = wrongTypeExceptionMap(propname)
    str = sprintf('''%s'' must be a map with keys of type char', propname);
    except = MException('matvhist:wrongType', str);
end


function [except] = wrongTypeExceptionPreviews(propname)
    str = sprintf('''%s'' must be a map with keys of type char', propname);
    except = MException('matvhist:wrongType', str);
end


function [except] = wrongTypeExceptionFile(propname)
    str = sprintf(['''%s'' must be an empty array or cell or of type ' ...
          'matvhist.VhistFile'], propname);
    except = MException('matvhist:wrongType', str);
end
