function [dict] = structToMap(aStruct, prefix)
% STRUCTTOMAP    Converts a matlab struct into a containers.Map object.
%    Returns a containers.Map object, which was created from a matlab
%    struct. A field in the struct can contain another struct (which causes
%    the method to recurse), a numeric or logical value, vector, matrix or
%    3d tensor, a string or cell of strings or a function handle.
%    All values are converted into strings, so that the returned dictionary
%    uses strings as keys and as values.
%
%    [DICT] = STRUCTTOMAP(ASTRUCT)
%
%    Converts ASTRUCT into a containers.map object with strings as keys
%    and values.
%
%    [DICT] = STRUCTTOMAP(ASTRUCT, PREFIX)
%
%    Converts ASTRUCT into a containers.map object with strings as keys
%    and values. All keys are prefixed with the string PREFIX, followed by
%    a '.'. structToMap(params, 'parameters') for example yields a map,
%    in which all keys are prefixed with 'parameters.'.

% init default prefix if neccessary
if nargin < 2
    prefix = '';
end

% dictionary, which will hold the results
dict = containers.Map;

fields = fieldnames(aStruct);

% handle all fields in the dictionary
for i = 1 : numel(fields)
    key = fields{i};
    prefixedKey = generateStructKey(prefix, key);
    value = aStruct.(key);
    
    dict2 = any2Dict(value, prefixedKey);
    dict = joinMaps(dict, dict2);
end

end


function [prefixedKey] = generateStructKey(prefix, key)

if strcmp(prefix, '')
    prefixedKey = key;
else
    prefixedKey = [prefix '.' key];
end

end


function [dict] = any2Dict(obj, prefix)

if ischar(obj)
    dict = containers.Map(prefix, sprintf('%s', obj));
elseif isa(obj, 'function_handle')
    dict = containers.Map(prefix, func2str(obj));
elseif isnumeric(obj) || islogical(obj)
    dict = tensor2Dict(obj, prefix);
elseif isstruct(obj) && numel(obj) > 1
    dict = containers.Map();
    for i = 1 : numel(obj)
        newDict = matvhist.structToMap(obj(i), sprintf('%s(%d)', prefix, i));
        dict = joinMaps(dict, newDict);
    end
elseif isstruct(obj)
    dict = matvhist.structToMap(obj, prefix);
elseif iscellstr(obj) && ndims(obj) <= 2 && (size(obj, 1) == 1 || size(obj, 2) == 1)
    dict = cellStr2Dict(obj, prefix);
end

end


function [dict] = cellStr2Dict(obj, prefix)

    dict = containers.Map();
    
    for i = 1 : numel(obj)
        key = [prefix sprintf('(%d)', i)];
        value = obj{i};
        dict(key) = value;
    end

end


function [dict] = tensor2Dict(obj, prefix)

if ndims(obj) <= 2
    dict = containers.Map(prefix, mat2str(obj));
elseif ndims(obj) <= 3
    dict = containers.Map();
    for i = 1:size(3)
        key = [prefix sprintf('(:, :, %d)', i)];
        dict(key) = mat2str(obj);
    end
else
    throw(MException('vhist:wrongDimension', 'Cannot convert tensors with more than 3 dimensions'));
end
end


function [joinedMaps] = joinMaps(map1, map2)

keys = map2.keys;
for i = 1 : numel(keys)
    key = keys{i};
    map1(key) = map2(key);
end
joinedMaps = map1;

end
