"""
Module containing everything needed to tokenize an input string via regular
expressions.



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: lexer.py 1071 2008-05-25 18:58:08Z ahuesgen $
"""

import re



class LexerError(RuntimeError):
    """
    An error which occured during lexing.

    """
    pass


class LexExpr(object):
    """
    An expression which is used by a Lexer. A lexer matches its input against
    the regex in this pattern and on success calls genFunction with id as second
    argument.
    
    """
    def __init__(self, regex, endToken, genFunction, id):
        """
        Creates a new lexer expression. LexExpr objects are used to tokenize the
        input string of the lexer.
        
        @type  regex: SRE_Pattern
        @param regex: A compiled regular expression used for matching.
                      Accessible through the member regex.
        
        @type  endToken: bool
        @param endToken: True if one occurence of this expression should stop
                         the Lexer. Accessible through the member endToken.
        
        @type  genFunction: function(matchresult, id)
        @param genFunction: Function, which is called when an occurence of this
                            token was found. The argument matchresult is a
                            regular expression search result and id is an int.
                            It should generate and return a token from the
                            matchresult or None if the token should be
                            discarded. If the returned value is not None, it is
                            added to the Token stream generated by the lexer. If
                            genFunction is None, all tokens generated by this
                            expression are discarded. Accessible through the
                            member genFunction.
        
        @type  id: int or None if genFunction is None
        @param id: The id passed as first argument to genFunction. Accessible
                   through the member id.
        
        """
        self.regex = regex
        self.endToken = endToken
        self.genFunction = genFunction
        self.id = id



class Lexer(object):
    """
    A lexer which tokenizes input strings.
    
    """
    def __init__(self):
        """
        Creates a lexer.
        
        """
        self.lexExprs = []
    
    def __reset(self):
        """
        Resets the lexer so that a new input string can be lexed.
        
        """
        self.tokens = []
    
    def addLexExpr(self, lexExpr):
        """
        Adds one LexExpr to the Lexer. These lexing expressions are used to
        find and process the single tokens in the input string. Note that the
        order, in which the LexExpr are added to the Lexer is the order, in
        which the LexExpr are applied to the input string. In some cases, it is
        important that one expression is checked before another one since the
        language described by one token may be a subset of or intersect with
        another token's language.
        
        @type  lexExpr: LexExpr
        @param lexExpr: The LexExpr object, which should be added to the Lexer.
        
        """
        self.lexExprs.append(lexExpr)
    
    def __handleToken(self, token):
        """
        Adds a token to the token stream list of the Lexer.
        
        @type  token: any
        @param token: The token, which is added to the token stream.
        
        """
        if token != None:
            self.tokens.append(token)
    
    
    def lexString(self, input):
        """
        Applies the Lexer to a string. If an error occures during lexing, a
        LexerError is raised. The resulting tokenstream can be
        received with getResult(), even if an error occured. So, if a token
        stores it's position, the position of the error can be determined by
        examining the last token in the unfinished stream.
        
        @type  input: string
        @param input: The string, which should be lexed and tokenized.
        
        """
        # reset the lexer
        self.__reset()
        foundEnd = False
        curPos = 0
        
        # tokenize the string
        while not foundEnd and curPos != len(input):
            match = None
            for expr in self.lexExprs:
                match = expr.regex.match(input, curPos)
                if match is not None:
                    if expr.genFunction is not None:
                        token = expr.genFunction(match, expr.id)
                        self.__handleToken(token)
                    foundEnd = expr.endToken
                    curPos += len(match.group(0))
                    break
            if match is None: # no LexExpr matched the input
                raise LexerError, "No Valid Token recognized (%s)" % (
                        repr(input[curPos:min(len(input)-1, curPos+10)]))
    
    def getResult(self):
        """
        Returns the tokenstream resulting from the previous lexing. The
        tokenstream is represented by a list of tokens.
        
        """
        return self.tokens
