"""
Module containing a parser which generates the layout for a pdf page.



This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

$Id: pdflayouter.py 2481 2012-04-12 14:36:52Z ahuesgen $
"""

import utilities, layoutlexer
from layoutlexer import ID_MARKER, ID_TEXT, ID_NEWLINE, ID_FORMAT

class LayoutError(Exception):
    """
    Exception, which is thrown if an error occures during pdf text layouting.
    """
    pass


def getNumLayoutLines(layout):
    numNewlines = 0
    for token in layout:
        if token["ID"] == ID_NEWLINE:
            numNewlines += 1
    return numNewlines + 1


class LayoutCursor(object):
    """
    Textcursor which records the (page, line, column) position within a text
    block.
    """
    def __init__(self, cols, lines):
        self.__curLine = 0
        self.__curPage = 0
        self.__cols = cols
        self.__lines = lines
        self.__lineSpace = self.__cols
    
    def progress(self, text):
        utilities.assertion(len(text) <= self.__lineSpace, "ERROR.....")
        self.__lineSpace -= len(text)
        if self.__lineSpace == 0:
            #self.newline()
            return True
        else:
            return False
    
    def newline(self):
        self.__lineSpace = self.__cols
        self.__curLine += 1
        if self.__curLine == self.__lines:
            self.newpage()
    
    def newpage(self, allowEmptyPages = False):
        if allowEmptyPages or self.__curLine != 0 or \
                self.__lineSpace != self.__cols:
            self.__curPage += 1
            self.__curLine = 0
            self.__lineSpace = self.__cols
    
    def curPageIsEmpty(self):
        return self.__curLine == 0 and self.__lineSpace == self.__cols
    
    def enoughSpace(self, text):
        return self.__lineSpace >= len(text)
    
    def fitsInEmptyLine(self, text):
        return len(text) < self.__cols
    
    def getLineNum(self):
        return self.__curLine
    
    def getPageNum(self):
        return self.__curPage
    
    def getLineSpace(self):
        return self.__lineSpace


class PDFLayouter(object):
    """
    A layout engine for text in a pdf page.
    """
    def __init__(self, cols, lines, supportMarkers):
        """
        Creates a new layouter.
        
        @type  cols: int
        @param cols: The maximum number of columns on a page.
        
        @type  lines: int
        @param lines: The maximum number of lines on a page.
        
        @type  supportMarkers: bool
        @param supportMarkers: True if the layouter should extract <id:x>
                               marker in the text. 
        """
        self.__cols = cols
        self.__lines = lines
        self.__markers = {}
        self.__lexer = layoutlexer.LayoutLexer(supportMarkers)
    
    def __createBlocks(self, text):
        def addBlock(blocks, newBlock):
            if newBlock != []:
                blockStr = "".join(newBlock)
                blocks.append(blockStr)
        
        lines = text.splitlines()
        
        addSpace = False
        blocks = []
        curBlock = []
        for line in lines:
            if line.strip() == "":
                curBlock.append("\n")
                addSpace = False
            elif line.strip() == "~":
                addBlock(blocks, curBlock)
                curBlock = []
                addSpace = False
            else:
                if addSpace:
                    line = " " + line
                curBlock.append(line)
                addSpace = True
        addBlock(blocks, curBlock)
        return blocks
    
    
    def __pageContainsWhitespaceOnly(self, page):
        for token in page:
            if token["ID"] == ID_TEXT:
                if token["TEXT"].strip() != "":
                    return False
        return True
     
    def __newTextToken(self, text, formatting):
        return {"ID": ID_TEXT, "TEXT": text, "FORMAT": formatting}
    
    def __newNewlineToken(self):
        return {"ID": ID_NEWLINE}
    
    def __newNewpageToken(self):
        return {"ID": ID_NEWPAGE,}
    
    def __handleNewPage(self, tokenList, oldPage, newPage):
        while oldPage < newPage:
            tokenList.append(self.__newNewpageToken())
            oldPage += 1
        return newPage
    
    def layoutText(self, text):
        """
        Layouts a given text. The text is broken into lines and pages, so
        that no line contains more then the given number of columns and no page
        contains more than the number of lines. Font types are applied, too.
        
        @type  text: unicode or str with one-byte-per-character-encoding
        @param text: The textblock, which should be formated.
        
        @rtype:  list of lists of dicts.
        @return: The formated text as a list of pages, each page comprising a
                 list of tokens. Each token is a dictionary with the fields
                 "ID" (values are ID_TEXT or ID_NEWLINE),
                 "FORMAT" (values are '_', '*' and None) and "TEXT"
                 (contains the formated text).
        """
        utilities.assertion(isinstance(text, (str, unicode)), "text is not "
                "of type str or unicode")
        
        # split text into blocks
        blocks = self.__createBlocks(text)
        
        # generate token streams for each block
        tokenBlocks = []
        for block in blocks:
            self.__lexer.lexString(block)
            tokenBlocks.append(self.__lexer.getResult())
        
        # generate formated strings
        formatedBlocks = []
        curFormatToken = None
        for block in tokenBlocks:
            curBlock = []
            for token in block:
                if token["ID"] == ID_TEXT:
                    token["FORMAT"] = curFormatToken
                    curBlock.append(token)
                elif token["ID"] in (ID_NEWLINE, ID_MARKER):
                    curBlock.append(token)
                elif token["ID"] == ID_FORMAT:
                    if curFormatToken == token["TYPE"]:
                        curFormatToken = None
                    elif curFormatToken == None:
                        curFormatToken = token["TYPE"]
                    else:
                        raise LayoutError, ("Found unexpected format token (%s "
                                "found, but %s expected)") % (token["TYPE"],
                                curFormatToken)
                else:
                    raise AssertionError, "Unreachable passage reached"
            formatedBlocks.append(curBlock)
        
        # layout text
        finishedText = []
        lastPageNum = 0
        nextIsNewline = False
        c = LayoutCursor(self.__cols, self.__lines)
        
        for block in formatedBlocks:
            curPage = []
            for token in block:
                if token["ID"] == ID_NEWLINE:
                    curPage.append(token)
                    c.newline()
                    nextIsNewline = False
                    if lastPageNum < c.getPageNum():
                        # add finished page
                        finishedText.append(curPage)
                        curPage = []
                        lastPageNum = c.getPageNum()
                
                elif token["ID"] == ID_MARKER:
                    self.__markers[token["VALUE"]] = (c.getPageNum(), c.getLineNum())
                    #print c.getPageNum(), c.getLineNum()
                
                elif token["ID"] == ID_TEXT:
                    restString = token["TEXT"]
                    format = token["FORMAT"]
                    while restString != "":
                        if nextIsNewline:
                            restString = restString.lstrip(" ")
                            if len(restString) != "":
                                curPage.append(self.__newNewlineToken())
                                c.newline()
                                nextIsNewline = False
                                if lastPageNum < c.getPageNum():
                                    # add finished page
                                    finishedText.append(curPage)
                                    curPage = []
                                    lastPageNum = c.getPageNum()
                            else:
                                continue
                        
                        if c.enoughSpace(restString): # token fits into line
                            curPage.append(self.__newTextToken(restString, format))
                            nextIsNewline = c.progress(restString)
                            restString = ""
                        else: # not enough space for complete string
                            pos = restString.rfind(" ", 0, c.getLineSpace()+1)
                            if pos >= 0: # word, which fits into line, found
                                tokenStart = restString[:pos]
                                if len(tokenStart) < c.getLineSpace():
                                    tokenStart += " "
                                restString = restString[pos+1:] # remove space
                                curPage.append(self.__newTextToken(tokenStart, format))
                                nextIsNewline = c.progress(tokenStart)
                            
                            else: # only one word (no spaces)
                                word = restString[0:(restString + " ").find(" ")]
                                if len(word) <= (self.__cols / 2) or c.getLineSpace() <= (self.__cols / 3):
                                    nextIsNewline = True
                                else: # word longer than one line
                                    tokenStart = restString[:c.getLineSpace()]
                                    restString = restString[c.getLineSpace():]
                                    curPage.append(self.__newTextToken(tokenStart, format))
                                    nextIsNewline = True
                        
                        if lastPageNum < c.getPageNum():
                            finishedText.append(curPage)
                            curPage = []
			lastPageNum = c.getPageNum()
            
            # test if last page must be added
            if not self.__pageContainsWhitespaceOnly(curPage):
                finishedText.append(curPage)
        
        return finishedText
    
    
    def genSimpleStream(self, text):
        return [self.__newTextToken(text, None)]
    
    def getMarkers(self, page):
        # has form [(id, (page, line)), ....]
        markerList = list(self.__markers.iteritems())
        return dict(filter(lambda x: x[1][0] == page, markerList))


'''
text = """Hello World Some String A quite Loooong Word """
#text = "short"

layouter = PDFLayouter(10, 3, False)
result = layouter.layoutText(text)
print result
'''

    
    
    
    
    
    
    
    

"""
        # []
        # debug output 
        print finishedText
        curLine = ""
        for token in finishedText:
            if token["ID"] == ID_NEWLINE:
                print curLine
                curLine = "" 
            elif token["ID"] == ID_TEXT:
                    curLine += token["TEXT"]
            elif token["ID"] == ID_NEWPAGE:
                print curLine
                curLine = ""
                print "------------------------------"

        print self.__markers
        """
        #return pages
'''
lo = PDFLayouter(30, 10, True)
x =  lo.layoutText(
"""Hello World

This is some text.
<id:2>foo bar Some more text hello argh foo bar... dummdidumm

1 2 3 *Hello World* _ARGH!!_FOOBAR<id:1>
~
A new page. Some Text.
""")

print x
print ""

print result
for page in result:
    curLine = ""
    for token in page:
        if token["ID"] == ID_TEXT:
            curLine += token["TEXT"]
        else:
            print curLine
            curLine = ""
    print curLine
    print "--------------------------------"
    curLine = ""
    
#print lo.getMarkers(0)
#print lo.getMarkers(1)
#print lo.getMarkers(2)

'''
"""*README*


This file conforms to the *VHIST* format developed at the Max-Planck-Institute 
for Neurological Research (MPIFNF), Cologne, Germany. 
VHIST, is an open specification, a detailed description and some related tools 
are available under the terms of an OpenSource license, 
http://www.nf.mpg.de/vhist. 
  Please see the LEGAL NOTICE below before proceeding.

  The general idea behind VHIST is to provide a robust and simple means for 
documenting a step of a workflow (e.g. 
quantification of a PET image volume, 
or (automated) image processing using SPM) by logging all relevant 
information: which files were used, which files were written,
what software 
package was used with what parameters. VHIST was conceived for usage on top of 
existing
workflows: ideally, you only have to add one line to an existing 
batch script.
  
VHIST files can act like a container for arbitrary chunks of information: you 
can embed log-files, (binary) image and header data. VHIST also conforms to 
the PDF-1.5 standard and can be used with standard PDF browsers (embedded data 
appears as embedded files in PDF browsers, each addition to an existing 
workflow
file strictly retains all previous information). 
  However, particular care was taken to facilitate extraction and processing 
of data (embedded files, in particular the automatically generated XML summary
of a workflow step) in an automated fashion and entirely independent of PDF 
specifics:
the human-readable summaries for each workflow step have only been 
provided as a convenience. A program written in the Python programming language
(xtract.py, 14 lines) has been attached to this document: it is sufficient 
(apart from a suitable Python distribution) to extract and validate all 
relevant information from this file. 


*LEGAL NOTICE*

This file may contain legally privileged and confidential information intended
solely for scientific use at the Max-Planck-Institute for Neurological 
Research, Cologne, Germany (MPINF). All data is property of the MPINF and may 
not be used in any way without prior written confirmation. 
  If you have received this file in error or if you are in doubt, please 
notify Email: security@nf.mpg.de."""

"""
Layout rules:
*text* ... text displayed in specialfont1 (text may contain newlines but no block boundaries)
_text_ ... text displayes in specialfont2 (text may contain newlines but no block boundaries)
text ~ text ... insert a newline at the given location

normal newlines in text are ignored.
an empty line forces a new paragraph (one empty line)
_ in ** blocks and * in __ blocks are ignored
**, __ and ~~ to escape *, _ and ~ in normal text

"""
