#!/usr/bin/env python
"""
This is the vhistadd commandline tool for creating VHIST files.


This file is part of the VHIST project, http://www.nf.mpg.de/vhist,
which defines a file format suitable for documenting workflows. VHIST
files can store meta-data about the workflow and allow to embed
arbitrary binary data, such as images or log files, inside the
documentation. Embedded data can be validated using multiple facilities.
VHIST files can be viewed with any PDF browser.

(c) Max-Planck-Institute for Neurological Research (MPIFNF)
Cologne, Germany 2005 - 2013
Email: vhist@nf.mpg.de, http://www.nf.mpg.de/vhist

This application is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
(LGPL version 2.1):

This application is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this application. If not, see
<http://www.gnu.org/licenses/>.

Author: Andreas Huesgen
$Id: vhistadd.py 908 2008-02-13 12:34:32Z ahuesgen $
"""
import sys, os

# load needed modules
from vhist import workflow, vhistconsts, pdfmain, utilities, \
        userinterface, namemodifier, logger, cmdparser

def readFiles(files, filetype):
    """
    Tries to load one of the files in the given filelist and returns the
    content. If no file could be loaded, None ist returned.
    
    @type  files: list of str
    @param files: A list of filenames. The function tries to load the files
                  one after the other until either one file was loaded
                  successfully or attempts for all files have failed.
    
    @type  filetype: str
    @param filetype: A string describing the type of the files in the list.
    
    @rtype:  str or None
    @return: The content of the loaded file or None if no file could be loaded
             successfully.
    """
    for filename in files:
        if filename != "":
            try:
                content = utilities.readFile(filename)
                logger.log(5, "successfully read %s file %s", filetype,
                        filename)
                return content
            except IOError, e:
                pass
    
    # no valid file found
    return None


def encodeStr(string):
    """
    Encodes the string with the default encoding and replaces unknown
    characters with a ? sign.
    
    @type  string: str or unicode
    @param string: The string to encode.
    
    @rtype:  str
    @return: The encoded string.
    """
    return string.encode(sys.getdefaultencoding(), "replace")


def initSystem(options):
    """
    Loads all needed config files (embedded readme, titlepage, critical file
    extensions).
    
    @type  options: optparse.Values
    @param options: Options from the commandline interface.
    
    @rtype:  (str, unicode)
    @return: The embedded readme and the content of the first page.
    """
    # common pathes
    join = os.path.join
    baseapp = os.path.realpath(sys.argv[0])
    basepath = os.path.dirname(baseapp)
    resPath = join(basepath, "../res/")
    commonPath = os.path.normpath(join(basepath, "../../vhist_common/"))
    vresPath = join(commonPath, "res/")
    
    readmeFiles = [options.readme,
            join(vresPath, "embedded_readme.txt"),
            join(resPath, "embedded_readme.txt")]
    
    firstpageFiles = [options.firstpage,
            join(vresPath, "title.txt"),
            join(resPath, "title.txt")]
    
    criticalFiles = [join(vresPath, "criticals.cfg"),
            join(resPath, "criticals.cfg")]

    # try to read a readme file
    readme = readFiles(readmeFiles, "readme")
    if readme == None:
        logger.abort(-1, "no readme file found")
    
    # try to read a first page file
    firstpage = readFiles(firstpageFiles, "first page")
    if firstpage == None:
        logger.abort(-1, "no first page file found")
    # try to convert readme file content from utf-8 to unicode
    try:
        firstpage = firstpage.decode("utf-8")
    except UnicodeDecodeError, e:
        logger.abort(-1, "The firstpage file %s is not encoded as utf-8", filename)
    
    # try to read critical extensions file
    critical = readFiles(criticalFiles, "criticals config")
    if critical == None:
        logger.abort(-1, "no criticals config file found")
    
    # parse critical files
    criticals = critical.splitlines()
    criticals = filter(lambda x: x != "" and x[0] != "#", criticals)
    
    # make critical extensions public
    namemodifier.setCriticalExtensions(criticals)
    
    return readme, firstpage


def logInformation(wf):
    """
    Logs information about the generated workflowstep.
    
    @type  wf: vhist.workflow.Workflow
    @param wf: The used workflow.
    """
    def __logFileInformation(f):
        # log information about one WorkflowFile f
        logger.log(4, "'%s'", encodeStr(f.filename))
        logger.log(5, "path: '%s'", encodeStr(f.filepath))
        logger.log(5, "filetype: '%s'", encodeStr(f.filetype))
        logger.log(5, "description: '%s'", encodeStr(f.description))
        logger.log(5, "embedded: %s", encodeStr(unicode(f.embedded)))
        if len(f.userAttrs) > 0:
            logger.log(5, "user attributes:")
            for key, value in f.userAttrs.iteritems():
                logger.log(5, "  <%s> = '%s'", encodeStr(key),
                        encodeStr(value))
        logger.log(5, "")
    
    # log config file/commandline interface configuration
    logger.log(3, "title: '%s'" % wf.title)
    logger.log(4, "timestamp: '%s'", encodeStr(wf.timestamp))
    logger.log(4, "description of workflow step: '%s'", encodeStr(wf.description))
    logger.log(4, "comment of workflow step: '%s'", encodeStr(wf.comment))
    logger.log(4, "executed command: '%s'", encodeStr(wf.command))
    logger.log(4, "software tool used: '%s'", encodeStr(wf.tool))
    logger.log(4, "path of tool '%s'", encodeStr(wf.toolPath))
    logger.log(4, "user: '%s'", encodeStr(wf.user))
    logger.log(4, "host: '%s'", encodeStr(wf.host))
    logger.log(4, "creator: '%s'", encodeStr(wf.creator))
    vfileList = ", ".join([encodeStr(f) for f in wf.genVhistFiles])
    logger.log(4, "output VHIST file(s): %s", vfileList)
    logger.log(4, "root VHIST file: %s", encodeStr(wf.vhistRootFile))
    
    # log user attributes
    if len(wf.userAttrs) > 0:
        logger.log(4, "user attributes:")
        for key, value in wf.userAttrs.iteritems():
            logger.log(4, "   <%s> = '%s'", encodeStr(key), encodeStr(value))
    
    # log information about files
    logger.log(4, "")
    logger.log(4, "input files:")
    for f in wf.files:
        if f.isInfile:
            __logFileInformation(f)
    
    logger.log(4, "")
    logger.log(4, "output files:")
    for f in wf.files:
        if f.isOutfile:
            __logFileInformation(f)


def initLogger(options):
    """
    Initializes the logging system to the requested verbosity.
    
    @type  options: optparse.Values
    @param options: Options from the commandline interface.
    """
    if options.quiet:
        logger.setLogLevel(logger.LOG_LEVEL_OFF)
    elif options.verbose:
        logger.setLogLevel(logger.LOG_LEVEL_MAX)
    else:
        logger.setLogLevel(3)


def findEncoding():
    """
    Finds and returns the default encoding of the system.
    
    @rtype:  str
    @return: A string containing the system's default encoding.
    """
    try:
        import locale
        enc = locale.getpreferredencoding()
    except:
        import sys
        enc = sys.stdin.encoding
    if enc != "":
        return enc
    else:
        # ah: fix for mac os x 10.5, which returns "" for some reason...
        return "ascii"


def decodeList(alist, encoding):
    return [arg.decode(encoding) for arg in alist]


def getEnvArgs():
    encoding = findEncoding()
    envArgs = []
    if "VHISTADD_ARGS" in os.environ.keys():
        envString = os.environ["VHISTADD_ARGS"]
        envArgs = decodeList(cmdparser.splitCommandline(envString), encoding)
    return envArgs


def assembleCommandline(args):
    encoding = findEncoding()
    envArgs = getEnvArgs()
    vhistaddName = args[0]
    args = [vhistaddName] + envArgs + args[1:]
    args = decodeList(args, encoding)
    
    return args


def readCommandfileOrStdin(commandfileOrStdin, firstCmdpath):
    if isinstance(commandfileOrStdin, (str, unicode)):
        # commandfile
        if firstCmdpath is None:
            firstCmdpath = os.path.dirname(commandfileOrStdin)
        logger.log(3, "reading commandfile '%s'..." % commandfileOrStdin)
        content = utilities.readFile(commandfileOrStdin)
    else:
        # stdin
        logger.log(3, "reading stdin...")
        content = commandfileOrStdin.read()
    return content, firstCmdpath


def readCommandline(args):
    encoding = findEncoding()
    envArgs = getEnvArgs()
    vhistaddName = args[0]
    
    # read initial commandline (commandline + env var)
    args = assembleCommandline(args)
    attributes, options = userinterface.parseCommandline(args)
    initLogger(options)
    
    logger.log(1, "This is %s." % (vhistconsts.getVhistCreatorVersion()))
    
    if options.stdin:
        options.cmdfiles.append(sys.stdin)
    
    # commandline contained arg files. Read these instead
    firstCmdpath = None
    if options.cmdfiles != []:
        try:
            logger.log(2, "found commandfile(s), all other options are ignored")
            args = [vhistaddName] + envArgs
            for cmdfile in options.cmdfiles:
                content, firstCmdpath = readCommandfileOrStdin(cmdfile, firstCmdpath)
                newArgs = cmdparser.splitCommandline(content, parseForVars=True)
                if len(newArgs) > 0:
                    args.extend(decodeList(newArgs, encoding))
            attributes, options = userinterface.parseCommandline(args, ignoreCmdFile=True)
        except Exception, e:
            if options.debug:
                raise
            else:
                logger.abort(-1, "An error occured while reading the "
                        "commandfile: '%s'", e)
    
    options.firstCmdpath = firstCmdpath
    return attributes, options


def syncCwd(options):
    # if the current working directory should be synced to a argfile
    # and there is an argfile, do so
    if options.syncCwd and options.firstCmdpath != "":
        try:
            firstCmdpath = os.path.realpath(options.firstCmdpath)
            logger.log(2, "changing current working directory to '%s'...", firstCmdpath)
            os.chdir(firstCmdpath)
        except Exception, e:
            if options.debug:
                raise
            else:
                logger.abort(-1, "An error occured while changing the current "
                        "working directory to '%s'", e)


def createWorkflow(attributes):
    # generate workflow step
    logger.log(2, "creating workflow step...")
    try:
        wf = workflow.Workflow()
        wf.genFromUI(attributes)
        return wf
    except workflow.WorkflowError, e:
        if options.debug:
            raise
        else:
            logger.abort(-1, "An error occured during the creation of the "
                "workflowstep: '%s'", e)


def appendWorkflowStep(wf, readme, firstpage, debugMode):
    if wf.vhistRootFile != u"":
        logger.log(2, "adding workflow step to VHIST file...")
        logger.log(3, "VHIST file: '%s'" % encodeStr(wf.vhistRootFile))
    try:
        pdfmain.addWorkflowStep(wf, readme, firstpage)
        logger.log(1, "workflow step added successfully.")
    except AssertionError, e:
        if debugMode:
            raise
        else:
            logger.abort(-1, "internal assertion failed: %s", e)
    except IOError, e:
        if debugMode:
            raise
        else:
            logger.abort(-1, "input/output error occured: %s", e)
    except Exception, e:
        if debugMode:
            raise
        else:
            logger.abort(-1, "error occured: %s", e)


def main(args):
    """
    Main function of vhistadd.
    """
    
    # read commandline (commandline + env var + arg files)
    attributes, options = readCommandline(args)
    
    # init special files and critical extenstions and change directory
    logger.log(2, "initialize...")
    readme, firstpage = initSystem(options)
    syncCwd(options)
    
    # create and log workflow step
    wf = createWorkflow(attributes)
    logInformation(wf)
    
    # append workflow step
    if not options.pretend:
        appendWorkflowStep(wf, readme, firstpage, options.debug)

main(sys.argv)

